@rem = ' This is a perl script for NT and UNIX.
@echo off
goto endofperl
@rem ';
# -*- tab-width: 4 -*-
#---------------------------------------------------------------------------+
# Licensed Materials - Property of IBM
# Rational ClearCase
# Copyright IBM Corp. 2000, 2013.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract
# with IBM Corp.
#---------------------------------------------------------------------------+

###########################################################################
# epoch_watchdog.bat
#
# This perl script is intended to be used regularly by MultiSite customers
# to check the self-consistency of the epoch tables in replicated VOBs.
# It may be invoked by the ClearCase Job Scheduler.
###########################################################################

use strict;
use Getopt::Long;

my $thisscript = 'epoch_watchdog';
my $tracing = ($ENV{TRACE_SUBSYS} && ($ENV{TRACE_SUBSYS} =~ m/$thisscript/));
my $verboseness = ($ENV{TRACE_VERBOSITY});
my $trace_msg ='';
my $error_msg = '';
my $logdir;
my $producthome;
my $error_logname = '';

use Config;

$CFG::including_script = $thisscript;

my $arch = $^O ? $^O : $main::Config{archname};
$CFG::NT = (($arch =~ /^MSWin32/) || ($arch eq 'i386-win32')) ? 1 : 0;

# Determine the correct path for including MScommon.pl
# CLEARCASEHOME e.v. should always be set on NT when called by ccperlpp.pl,
# so exit if it's not set. On UNIX, assume /opt/rational/clearcase if user
# hasn't explicitly set CLEARCASEHOME (or deprecated ATRIAHOME).
#
if ($CFG::NT) {
    #  It's NT
    if (!exists $ENV{CLEARCASEHOME} ) {
	print STDOUT "CLEARCASEHOME environment variable not set\n";
	exit 1;
    }
    $producthome = $ENV{CLEARCASEHOME};
    chdir $producthome;
    #  For CC V3.2.1 only, uncomment the following line:
    #push(@INC, "$producthome\\config\\cron");
    #  For CC V4.0 and above, uncomment the following line:
    push(@INC, "$producthome\\config\\scheduler\\tasks");
    $logdir = "$producthome\\var\\log\\epoch_logs";
} else {
    #  It's UNIX
    # Check that root is running the program; die otherwise
    my $user = getpwuid ($>);
    if ( $user ne "root" ) {
    	print STDOUT "Must be root to execute epoch_watchdog script\n";
	exit 1;
    }

    if (!exists $ENV{CLEARCASEHOME}) {
        # check for (now deprecated) ATRIAHOME e.v.; use if set.
        if (exists $ENV{ATRIAHOME} ) {
            $ENV{CLEARCASEHOME} = $ENV{ATRIAHOME};
        } else {
            $ENV{CLEARCASEHOME} = "/opt/rational/clearcase";
        }
    }
    $producthome = $ENV{CLEARCASEHOME};

    chdir $producthome;
    #  For CC V3.2.1 only, uncomment the following line:  
    #push(@INC, "$producthome/config/multisite");
    #  For CC V4.0 and above, uncomment the following line:
    push(@INC, "$producthome/config/scheduler/tasks");
    $logdir = "/var/adm/rational/clearcase/log/epoch_logs";
}

# set script version to test for mismatch against the MScommon.pl version.
$CFG::MS_script_version = "7.0.0.0";

require 'MScommon.pl';
require 5.8.6;

if (!$CFG::MScommon_version) {
    $CFG::MScommon_version = "(unknown older version)";
}

# check for mismatched versions - MScommon.pl should set
# $CFG::MScommon_version; if not set or unequal, may indicate that an
# older/newer version is being included improperly.

if ($CFG::MScommon_version ne $CFG::MS_script_version) {
    die "Version mismatch between require file 'MScommon.pl' " .
        "(version: ${CFG::MScommon_version}) and \n" .
        "the MultiSite script which includes it, 'epoch_watchdog.bat' " .
        "(version: ${CFG::MS_script_version}).\n" . 
        "Ensure that old versions of the script/require file are not in use.\n";
}

$CFG::failcount = 0;

if ($tracing) {
    $| = 1 ;  # do not buffer stdout
}	

  traceprint ("running on architecture $arch\n", 4);

###############################################################################
# Main 
#

# Parse the command line options:

    traceprint ("ARGV: " . join(' ',@ARGV) . "\n", 1);

my $VOBtag = '';
my $help = 0;
my $all_local = 0;
my $VOBs;
my $listfile = '';
my $usage = 
    "$thisscript  [-h] [-all | -vobs <VOB-tag, ...> | list_file ]\n" .
    "    More help is available by using the '-help' command-line switch.\n";
my $helpmsg = <<ENDHELP;
================================================================================

"epoch_watchdog" monitors one or more VOB replicas for oplog divergence, by 
determining if the replica's epoch numbers have rolled back WITHOUT a 
restorereplica command having been run. This situation could imply that a 
replicated VOB was improperly recovered from backup. 


$usage
Command line options:
    -help               Prints this message and exits.
    
    -all                Check all local replicated VOBs. Other "-vob"
                        parameters on command line and any listfile specified
                        will be ignored if '-all' is specified.

    -vobs               Full VOB-tag,  in comma-separated list. No whitespace.

    list_file           Path to file containing a set of directives to 
                        execute. Must be the final parameter specified on
                        the command line. (Default is no list-file).

                        List-file may contain the following directive:
                        vob:<VOBtag1> 
                        vob:<VOBtag2>
			... 
                        vob:<VOBtagN> 
 			One entry to a line, no whitespace

================================================================================
ENDHELP

my $st = GetOptions("help" => \$help,
                "all" => \$all_local,
		"vobs=s" => \$VOBs);

if ($help) {
    print STDOUT "$helpmsg\n";
    exit 0;
}

if (!($all_local) && !($VOBs) && (scalar @ARGV == 0)) {
    print STDOUT "$usage\n";
    exit 0;
}

# Enable override to be sure we can open all local VOBs.
$ENV{"CCASE_ACL_RECOVERY_MODE"} = "QUIET";

#
# Validate that the log directory actually exists and that we can write to it.
#

my $log_status = CheckLogDir($logdir);

if ($log_status != 0) { 

    if($log_status == 1){
    	$error_msg = "ERROR: Unable to create log directory $logdir\n";
    }elsif ($log_status == 2){
    	$error_msg = "ERROR: Log directory not writeable: $logdir\n";
    }else {
    	$error_msg = "ERROR: Unknown error checking $logdir\n";
    }

    print STDOUT $error_msg;
    exit 1;
}
		

#
# Redirect STDERR to a uniquely named log file
#

$error_logname = RedirectSTDERRtoLog ($logdir, 'epoch');

if (scalar @ARGV > 1) {
    $error_msg = "ERROR: too many list-files specified in cmd line!\n";
    WriteLog ($error_logname, $error_msg);
    LogError ($thisscript, $error_logname);
    print STDOUT $error_msg;
    exit 1;
}

if (!$st) {  # bad status or more than one arg left
    $error_msg = "ERROR: error getting command line options.\nUsage:\n$usage\n";
    WriteLog ($error_logname, $error_msg);
    LogError ($thisscript, $error_logname);
    print STDOUT $error_msg;
    exit 1;
}

if (scalar @ARGV) {  # list-file specified
    $listfile = $ARGV[0];
    chomp $listfile;
} 

if ( $tracing) {
    $trace_msg = "\n===============================================".
    	"=============\n" .
    	"Beginning execution of MultiSite script '$thisscript' at " .
    	MakeTimestamp() . ".\n\n".
    	"command options specified or defaulted:\n  logdir: $logdir\n".	
    	"  all: $all_local\n    vobs: $VOBs\n".
	"  list: $listfile\n\n".
	"error log filename: $error_logname\n";

	WriteLog ("$trace_msg\n");
        $CFG::failcount++;
}
traceprint ($trace_msg, 1);

#
# Get the leaf name(s) of the VOB(s).

my @vobs;
my $tag = '';

#
# if -all specified, ignore -vob and listfile

if ($all_local) {
    @vobs = GetLocalVobs();
    $VOBtag = '';
    $listfile = '';
} 

#
# If -vobs, separate list into an array of VOB-tags. Ignore listfile

if ($VOBs) {
    @vobs  = split /,/,$VOBs;
    chomp @vobs;
    $listfile = '';
}


#
# If listfile specified, make sure it exists

if ($listfile && !(-e $listfile )) {
    $error_msg = "ERROR: List file $listfile does not exist\n";
    WriteLog ($error_logname, $error_msg);
    LogError ($thisscript, $error_logname);
    print STDOUT $error_msg;
    exit 1;
}

#
# It exists. Try to read the list of VOB-tags

if ($listfile) {
    my ($option, $val);
    traceprint ("opening list file '$listfile'.\n", 2);
    if ( open (SCR, "<$listfile")){
    	while (<SCR>) {
            chomp;
            ($option, $val) = split(":", $_);
            traceprint ("listfile - option: $option value: $val\n",2);
            if ("vob" =~ m/^$option/) {

        	@vobs = ($val, @vobs);
            	traceprint ("found vob...$val\n",2);

	    } else {

		$error_msg = "ERROR: unrecognized option '$option' in listfile $listfile\n";
		WriteLog($error_logname, $error_msg);
    		LogError ($thisscript, $error_logname);
		print STDOUT $error_msg;
        	$CFG::failcount++;
            }
    	}
    } else {
	# Couldn't open the list file. Output error and exit.
	$error_msg = "ERROR: Could not open $listfile for reading\n";
	WriteLog ($error_logname, $error_msg);
    	LogError ($thisscript, $error_logname);
    	print STDOUT $error_msg;
    	exit 1;
    }
	
}

my $res = 0;
my @leaf;
my $recovering = 0;
my $VOBEpochNumber = 0;
my $VOBleaf = '';
my $EpochLogName = '';
my $EpochLogFile = '';
my $LogVal = 0;
my $replica_oid = '';
my $replica_name = '';
my @siblings;
my $export_day;
my $export_month;

LOOP: foreach $tag (@vobs) {
    
    # First, get the replica name from the VOB-tag

    if ($CFG::NT){
	@leaf = split /\\/,$tag;
    } else {
	@leaf = split /$CFG::FS/, $tag;
    }

    $VOBleaf = pop(@leaf);

    $replica_name = GetThisReplicaName ( $tag );
    if ($replica_name eq "unreplicated") {
	next LOOP;
    }
    if($? || !($replica_name)) {
	$error_msg = "epoch_watchdog: ERROR: Could not get replica name for VOB $tag.\n";
	WriteLog ($error_logname, $error_msg);
    	LogError ($thisscript, $error_logname);
    	print STDOUT $error_msg;
        $CFG::failcount++;
	next LOOP;
    }	

    # Now, get the replica_oid, to identify the proper epoch number log file

    $replica_oid = GetOid ( "replica:${replica_name}\@${tag}" );
    if($?) {
	$error_msg = "epoch_watchdog: ERROR: Could not get oid for replica $replica_name.\n";
	WriteLog ($error_logname, $error_msg);
    	LogError ($thisscript, $error_logname);
    	print STDOUT $error_msg;
        $CFG::failcount++;
	next LOOP;
    }

    #
    # Get this replica's epoch number from the VOB database, using multitool subcommands

    $VOBEpochNumber = MultitoolLsepoch ( $tag, $replica_name, $replica_oid );
    if($?) {
	$error_msg = "epoch_watchdog: ERROR: Couldn't get epoch number for $replica_name\@$tag from database.\n";
	WriteLog ($error_logname, $error_msg);
    	LogError ($thisscript, $error_logname);
    	print STDOUT $error_msg;
        $CFG::failcount++;
	next LOOP;
    }

    #
    # Parse the basename of the epoch number log file.

    my $rep_oid_dense = $replica_oid;
    $rep_oid_dense =~ tr/.://d;
    $EpochLogName = $VOBleaf . "." . $rep_oid_dense;
    $EpochLogFile = "$logdir${CFG::FS}$EpochLogName";

    $trace_msg = "epoch log filename: $EpochLogFile\n";
    if ( $tracing ) {
	WriteLog($error_logname, $trace_msg);
        $CFG::failcount++;
    }
    traceprint ($trace_msg, 2);

    #
    # Open the epoch number log file and retrieve the value stored there
    # Create the file if it doesn't already exist

    if ( open (EPOCHFILE, "<$EpochLogFile") ){

    	# Epoch Log File exists. Get the $LogVal

    	$LogVal = <EPOCHFILE>;
    	chomp $LogVal;
        close(EPOCHFILE);

    } else { 

	# It doesn't exist. Create it, and store $VOBEpochNumber

    	$LogVal = $VOBEpochNumber;
	my $msg = "epoch_watchdog: Creating $EpochLogFile with value $LogVal.\n";
        WriteLog ($error_logname, $msg);

    	traceprint ($msg, 1);
        $CFG::failcount++;
    	
	if (!(open (EPOCHFILE, ">$EpochLogFile"))){
	    my $error_msg = "epoch_watchdog: ERROR: Can't write to $EpochLogFile.\n";
	    WriteLog ($error_logname, $error_msg);
    	    LogError ($thisscript, $error_logname);
    	    print STDOUT $error_msg;
            $CFG::failcount++;
	}
    	print EPOCHFILE "$VOBEpochNumber\n";
        close(EPOCHFILE);
	next LOOP;
    }
 
    # The real work: Compare the VOB's epoch number to the log's value

    if ( ($LogVal < $VOBEpochNumber) ) {

	# Conditions normal. Record the new value.

    	if ( !(open (EPOCHFILE, ">$EpochLogFile"))) {
	    $error_msg = "Can't write to $EpochLogFile.\n";
	    WriteLog ($error_logname, $error_msg);
    	    LogError ($thisscript, $error_logname);
	    print STDOUT $error_msg;
	    exit 1;
	}
    	print EPOCHFILE "$VOBEpochNumber\n";
    	close(EPOCHFILE);
    }

    if ( $LogVal > $VOBEpochNumber ) {

    	# Should only occur if replica is undergoing restorereplica--check!
    	
    	$recovering = GetRestoreFlag ( $tag );
    	if($?) {
	    $error_msg = "epoch_watchdog: ERROR: Couldn't get flags for $replica_name\@$tag.\n";
	    WriteLog ($error_logname, $error_msg);
    	    LogError ($thisscript, $error_logname);
    	    print STDOUT $error_msg;
            $CFG::failcount++;
    	}

    	if ( $recovering == 1 ) {

            #
            # undergoing restorereplica, remind the customer to remove the 
            # current $EpochLogFile when the restore process is completed.

	    my $msg = "epoch_watchdog: WARNING: Replica $replica_name\@$tag is undergoing a restorereplica process.\nWhen that process completes, and you receive the message that it is safe to unlock the VOB,\nyou must delete $EpochLogFile to resume normal operation using epoch_watchdog.\n";
            WriteLog ($error_logname, $msg);
    	    LogError ($thisscript, $error_logname);
    	    print STDOUT $msg;
            $CFG::failcount++;
    	} else {

    	    #
    	    # Not in restorereplica process: Lock the VOB and sound the alarms!

	    $error_msg = "DANGER! Possible divergence detected by epoch_watchdog".
	        " script in $tag.\n" .
	        "Log Value is $LogVal; VOB Epoch Number is $VOBEpochNumber.\n" . 
	        "Log is $EpochLogFile.\n";

	    WriteLog ($error_logname, $error_msg);
    	    LogError ($thisscript, $error_logname);
    	    print STDOUT $error_msg;
    	    $CFG::failcount++;
	
	    $res = CleartoolLockVOB ( $tag );
    	    if($res) {
	 	$error_msg = "epoch_watchdog: ERROR: Could not lock VOB $tag.\n";
	        WriteLog ($error_logname, $error_msg);
    	    	LogError ($thisscript, $error_logname);
    		print STDOUT $error_msg;
    	        $CFG::failcount++;
    	    } else {
	    	my $msg = "VOB $tag locked by epoch_watchdog script at " . 
 		MakeTimestamp() . ".\n";
		WriteLog ($error_logname, $msg);
    	    	traceprint ($msg, 1);
	        $CFG::failcount++;
	    }

	    # Try to find out when local site sync'ed to its siblings 

    	    # What are its siblings' names?

    	    @siblings = GetSibs ( $tag );
    	    my $sib;
    	    if($?) {
		$error_msg = "epoch_watchdog: ERROR: Couldn't get siblings for $replica_name\@$tag.\n";
	        WriteLog ($error_logname, $error_msg);
    	    	LogError ($thisscript, $error_logname);
    		print STDOUT $error_msg;
	        $CFG::failcount++;
    	    }

    	    #
    	    # Get the lshistory for each sibling. Report most recent exports to that site.
    	    # Default will be cleartool lshistory -since <forever>

    	    foreach $sib (@siblings) {
		my $msg = '';
     		my $export_history = GetHistory ( $sib, $tag );
     	    	if($?) {
		    $error_msg = "epoch_watchdog: ERROR: Couldn't get siblings for $replica_name\@$tag.\n";
	            WriteLog ($error_logname, $error_msg);
    	    	    LogError ($thisscript, $error_logname);
    		    print STDOUT $error_msg;
	            $CFG::failcount++;
    	    	}

   		if(!$export_history) {
		    $msg = "\nepoch_watchdog: WARNING: lshistory shows no exports from replica:$replica_name\@$tag to replica:$sib.\n";
		} else {
    		    $msg = "Export history:Exported to replica $sib on $export_history\n".
			"Do cleartool lshistory replica:$sib\@$tag for more information\n";
		}
		WriteLog ($error_logname, $msg);
    		print STDOUT $msg;
	        $CFG::failcount++;
	    }
    	}
    }
}

my $msg = "Successfully completed $thisscript script\n";
if ($tracing) {
	WriteLog ($error_logname, $msg);
        $CFG::failcount++;
}
traceprint ($msg, 1);

print STDOUT "";

# Remove empty logfile if no errors were encountered:

close STDERR;
if (!$CFG::failcount) {
	unlink ($error_logname) || warn "Unable to delete $error_logname\n";
}

close STDOUT;

######################## SUBROUTINES #########################################
#
# Describe the VOB to get the replica name
sub GetThisReplicaName {
    my ($Vobtag) = @_;
    my $rep_name = '';
    my $fmt_qq = '';

    #First, check to see if this VOB is replicated:
    if ($CFG::NT){
	$fmt_qq = qq/"%[vob_replication]p\\n"/;
    } else {
	$fmt_qq = qq/"%[vob_replication]p\n"/;
    }

    my $cmd = "$CFG::ClearTool desc -fmt ${fmt_qq} vob:${Vobtag}";

    my $is_replicated = DoCmdNoTranslate ($cmd);

    chomp $is_replicated;

    if ($is_replicated eq "replicated"){    
    	if ($CFG::NT){
    		$fmt_qq = qq/"%[replica_name]p\\n"/;
    	} else {
    		$fmt_qq = qq/"%[replica_name]p\n"/;
    	}
    	$cmd = "$CFG::ClearTool desc -fmt ${fmt_qq} vob:${Vobtag}";
    	$rep_name = DoCmd ($cmd);
      	chomp $rep_name;
    } else {
	$rep_name = "unreplicated";
    }	

    return $rep_name;
}

#
# Get this VOB's lsepoch, and retrieve the replica's epoch number
sub MultitoolLsepoch {
    my ($Vobtag, $rep_name, $rep_oid) = @_;
    my $VOBnumber = 0;
    my $j = 0;
    my $start = 0;

    my $cmd = qq($CFG::MultiTool lsepoch $rep_name\@$Vobtag);
    traceprint ("lsepoch cmd: $cmd\n", 2);

    my $lsepoch_output = DoCmdNoTranslate ($cmd);

    my @array = split /^/,$lsepoch_output;
    chomp @array;
    traceprint (@array . "\n", 4);

    #
    # Get array length
    my $length = $#array +1;
    my $found_it = 0;

    for ( $j = $start; $j < $length; $j++ ){
	next unless $array[$j] =~
	    m@^(\s+)oid:$rep_oid=(\d+)@m;
	$VOBnumber = $2;
	chomp $VOBnumber;
	$found_it = 1;
	last;
    }
    
    if (! $found_it) {
	return 1;
    }

    return $VOBnumber;
}

#
# Use lsreplica -long to see if this replica is in a restorereplica process
sub GetRestoreFlag {
    my ($Vobtag) = @_;
    my $line = '';
    my $is_recovering = 0;

    my $cmd = qq($CFG::MultiTool lsreplica -long -invob $Vobtag);
    traceprint ("lsreplica cmd: $cmd\n", 2);

    my $lsreplica_output = DoCmdNoTranslate ($cmd);

    my @array = split /^/,$lsreplica_output;
    chomp @array;
    traceprint (@array . "\n", 4);

    foreach $line (@array){
	next unless $line =~
	    m@^(\s+)replica is recovering@m;
	$is_recovering = 1;
	chomp $line;
	last;
    }
    
    return $is_recovering;
}

#
# Return list of siblings in VOB family
sub GetSibs {
    my ($Vobtag) = @_;

    my $cmd = qq($CFG::MultiTool lsreplica -short -siblings -invob $Vobtag);
    traceprint ("lsreplica cmd: $cmd\n", 2);
    my $sibs = DoCmd ($cmd);
    my @Siblings = split /^/, $sibs;
    chomp @Siblings;
    traceprint (@Siblings . "\n", 4);
    return @Siblings;
}

#
# Return export lshistory for a sibling in a VOB family
sub GetHistory {
    my ($Sib, $VOBtag) = @_;
    my $histline = '';
    my $export_day = 0;
    my $export_month = 0;
    my $export_hist = '';

    my $cmd = qq($CFG::ClearTool lshistory replica:$Sib\@$VOBtag);
    traceprint ("lshistory cmd: $cmd\n", 2);

    my $lsh_out = DoCmdNoTranslate ($cmd);

    my @array = split /^/, $lsh_out;
    chomp @array;
    traceprint ("Export history to $Sib\@$VOBtag:\n", 3);
    traceprint (@array . "\n", 3);
    foreach $histline (@array){
	# The new lshistory output has the form
	# "2005-4-29T08:24 username export sync from replica X to replica Y"
	# and we'd like to grab the day & month for our output
    	next unless $histline =~
		m@^(.*)(\d{2})-(\d{2})T(..):(.*)(\s*)export (.*)$@m;
		$export_day = $3;
		$export_month = $2;
		$export_hist = "$export_day-$export_month";
		last;
    }

    return $export_hist;
}

#
# Create a local (non-oplogged) lock on the VOB 
sub CleartoolLockVOB {
    my ($Vobtag) = @_;

    my $status = system (qq($CFG::ClearTool lock vob:$Vobtag));
    traceprint ("lock cmd: $CFG::ClearTool lock vob:$Vobtag\n", 1);
    if ($status) {
	return 1;
    }

    return 0;
}


__END__
:endofperl
ccperl -S ccperlpp.pl %0.bat %*
